function [Kuu, Kuv, Kvu, Kvv] = K_kernel_solver(N, a, b, gamma, q, numIter)
%% [Kvu, Kvv] = K_kernel_solver_final(N, a, b, gamma, q, numIter)
% Solves a system on the form
%   Kuu_x(x, xi) - Kuu_xi(x, xi) = - b exp(-gamma xi) Kuv(x, xi)
%   Kuv_x(x, xi) + Kuv_xi(x, xi) = - a exp( gamma xi) Kuu(x, xi)
%   Kvu_x(x, xi) - Kvu_xi(x, xi) =   b exp(-gamma xi) Kvv(x, xi)
%   Kvv_x(x, xi) + Kvv_xi(x, xi) =   a exp( gamma xi) Kvu(x, xi)
%   
%   with  BCs:
%   Kuu(x, 0) = (1/q) Kuv(x, 0)
%   Kvv(x, x) =  0.5 a exp( gamma x)
%   Kvu(x, x) = -0.5 b exp(-gamma x)
%   Kvv(x, 0) = q Kvu(x, 0)
%
%   using the method of Characteristics with numIter iterations. The result
%   is returned in two N x N matrices.


    %% Implementation starts here
    Delta = 1/(N - 1);

    mask = triu(ones(N));

    v0 = mask * (0.5*a);
    w0 = mask * (-0.5*b);
    
    u = w0;
    v = q*w0;
    w = w0;
    z = q*w0;

    for iter = 1:numIter
        %% I_w1
        I_v1 = zeros(N);
        I_w1 = zeros(N);
        
        % Special treatment for the first diagonal line
        for j = 0:(N - 2)
            I_v1(j + 1, j + 2) = ...
                + 0.25*u(j + 1, j + 1) ...
                + 0.25*u(j + 2, j + 2) ...
                + 0.5*u(j + 1, j + 2);
            
            I_w1(j + 1, j + 2) = ...
                + 0.25*z(j + 1, j + 1) ...
                + 0.25*z(j + 2, j + 2) ...
                + 0.5*z(j + 1, j + 2);
        end;

        for i = 2:1:(N - 1)
            for j = 0:(N - 1 - i)
                I_v1(j + 1, i + j + 1) = ...
                    + I_v1(j + 2, i + j) ...
                    + 0.25*u(j + 1, i + j) ... 
                    + 0.75*u(j + 1, i + j + 1) ...
                    + 0.75*u(j + 2, i + j) ...
                    + 0.25*u(j + 2, i + j + 1);
                
                I_w1(j + 1, i + j + 1) = ...
                    + I_w1(j + 2, i + j) ...
                    + 0.25*z(j + 1, i + j) ... 
                    + 0.75*z(j + 1, i + j + 1) ...
                    + 0.75*z(j + 2, i + j) ...
                    + 0.25*z(j + 2, i + j + 1);
            end;
        end;


        %% I_z1
        I_u1 = zeros(N);
        I_z1 = zeros(N);

        for i = 1:(N - 1)
            dummy_u = 0;
            dummy_z = 0;

            for k = 0:i
                if ((k == 0) || (k == i))
                    sigma = 0.5;
                else
                    sigma = 1;
                end;

                if (mod(i + k, 2))
                    if (k == 0)
                        dummy_u = dummy_u + sigma * 0.5*( ...
                                u(1 + round(0.5*(i - k + 1)), 1 + round(0.5*(i + k + 1))) ... 
                              + u(1 + round(0.5*(i - k - 1)), 1 + round(0.5*(i + k - 1))));
                          
                        dummy_z = dummy_z + sigma * 0.5*( ...
                                z(1 + round(0.5*(i - k + 1)), 1 + round(0.5*(i + k + 1))) ... 
                              + z(1 + round(0.5*(i - k - 1)), 1 + round(0.5*(i + k - 1))));
                    else
                        dummy_u = dummy_u + sigma * 0.25*( ...
                                u(1 + round(0.5*(i - k + 1)), 1 + round(0.5*(i + k + 1))) ... 
                              + u(1 + round(0.5*(i - k + 1)), 1 + round(0.5*(i + k - 1))) ...
                              + u(1 + round(0.5*(i - k - 1)), 1 + round(0.5*(i + k + 1))) ...
                              + u(1 + round(0.5*(i - k - 1)), 1 + round(0.5*(i + k - 1))));
                          
                        dummy_z = dummy_z + sigma * 0.25*( ...
                                z(1 + round(0.5*(i - k + 1)), 1 + round(0.5*(i + k + 1))) ... 
                              + z(1 + round(0.5*(i - k + 1)), 1 + round(0.5*(i + k - 1))) ...
                              + z(1 + round(0.5*(i - k - 1)), 1 + round(0.5*(i + k + 1))) ...
                              + z(1 + round(0.5*(i - k - 1)), 1 + round(0.5*(i + k - 1))));
                    end;
                else
                    dummy_u = dummy_u + sigma * ...
                          u(1 + round(0.5*(i - k)), 1 + round(0.5*(i + k)));
                      
                    dummy_z = dummy_z + sigma * ...
                          z(1 + round(0.5*(i - k)), 1 + round(0.5*(i + k)));
                end;
            end;

            I_u1(1, i + 1) = dummy_u;
            I_z1(1, i + 1) = dummy_z;
        end;
        
        
        % Create Toeplitz matrix
        for i = 1:N
            I_u1(i, i:N) = I_u1(1, 1:(N - i + 1)); % Faster than toepelitz()
            I_z1(i, i:N) = I_z1(1, 1:(N - i + 1));
        end;


        %% I_z2
        I_u2 = zeros(N);
        I_z2 = zeros(N);

        for i = 1:(N - 1)
            for j = 1:i
                I_u2(j + 1, i + 1) = I_u2(j, i) + 0.5*(v(j, i) + v(j + 1, i + 1));
                I_z2(j + 1, i + 1) = I_z2(j, i) + 0.5*(w(j, i) + w(j + 1, i + 1));
            end;
        end;
        
        
        %% Store values, and start over again
        u =(1/q)*v0 - (1/q)*a*0.5*Delta*I_u1 - b*Delta*I_u2;
        v = v0 - 0.5*a*Delta*I_v1;
        w = w0 + 0.5*b*Delta*I_w1;
        z = q*w0 + q*b*0.5*Delta*I_z1 + a*Delta*I_z2;
    end;

    grid = kron((0:Delta:1), ones(N, 1));
    e_plus  = exp(0.5*gamma*grid);
    e_minus = exp(-0.5*gamma*grid);

    Kuu = e_plus.*e_minus'.*u;
    Kuv = e_plus.*e_plus'.*v;
    Kvu = e_minus.*e_minus'.*w;
    Kvv = e_minus.*e_plus'.*z;