function [Laa, Lab, Lba, Lbb] = L_kernel_solver(N, a, b, gamma, q, numIter)
%% [Kvu, Kvv] = K_kernel_solver_final(N, a, b, gamma, q, numIter)
% Solves a system on the form
%   Kvu_x(x, xi) - Kvu_xi(x, xi) = a Kvv(x, xi)
%   Kvv_x(x, xi) + Kvv_xi(x, xi) = b Kvu(x, xi)
%   
%   with  BCs:
%   Kvu(x, x) = -0.5 c_2 exp(-gamma x
%   Kvv(x, 0) = q Kvu(x, 0)
%
%   using the method of Characteristics with numIter iterations. The result
%   is returned in two N x N matrices.


    %% Implementation starts here
    Delta = 1/(N - 1);

    mask = triu(ones(N));
    
    v0 = mask * (0.5*a);
    w0 = mask * (-0.5*b);

    qi = 1/q;
    
    u = qi*v0;
    v = v0;
    w = w0;
    z = q*w0;

    for iter = 1:numIter
        %% I1
%         tic;
        I_v1 = zeros(N);
        I_w1 = zeros(N);
    
        % Special treatment for the first diagonal line
        for j = 0:(N - 2)
            I_v1(j + 1, j + 2) = ...
                + 0.25*z(j + 1, j + 1) ...
                + 0.25*z(j + 2, j + 2) ...
                + 0.5*z(j + 1, j + 2);
            
            I_w1(j + 1, j + 2) = ...
                + 0.25*u(j + 1, j + 1) ...
                + 0.25*u(j + 2, j + 2) ...
                + 0.5*u(j + 1, j + 2);
        end;

        for i = 2:1:(N - 1)
            for j = 0:(N - 1 - i)
                I_v1(j + 1, i + j + 1) = ...
                    + I_v1(j + 2, i + j) ...
                    + 0.25*z(j + 1, i + j) ... 
                    + 0.75*z(j + 1, i + j + 1) ...
                    + 0.75*z(j + 2, i + j) ...
                    + 0.25*z(j + 2, i + j + 1);
                I_w1(j + 1, i + j + 1) = ...
                    + I_w1(j + 2, i + j) ...
                    + 0.25*u(j + 1, i + j) ... 
                    + 0.75*u(j + 1, i + j + 1) ...
                    + 0.75*u(j + 2, i + j) ...
                    + 0.25*u(j + 2, i + j + 1);
            end;
        end;
%         toc;


        %% I2
%         tic;
        I_u1 = zeros(N);
        I_z1 = zeros(N);

        for i = 1:(N - 1)
            dummy_u = 0;
            dummy_z = 0;

            for k = 0:i
                if ((k == 0) || (k == i))
                    sigma = 0.5;
                else
                    sigma = 1;
                end;

                if (mod(i + k, 2))
                    if (k == 0)
                        dummy_u = dummy_u + sigma * 0.5*( ...
                                z(1 + round(0.5*(i - k + 1)), 1 + round(0.5*(i + k + 1))) ... 
                              + z(1 + round(0.5*(i - k - 1)), 1 + round(0.5*(i + k - 1))));
                        dummy_z = dummy_z + sigma * 0.5*( ...
                                u(1 + round(0.5*(i - k + 1)), 1 + round(0.5*(i + k + 1))) ... 
                              + u(1 + round(0.5*(i - k - 1)), 1 + round(0.5*(i + k - 1))));
                    else
                        dummy_u = dummy_u + sigma * 0.25*( ...
                                z(1 + round(0.5*(i - k + 1)), 1 + round(0.5*(i + k + 1))) ... 
                              + z(1 + round(0.5*(i - k + 1)), 1 + round(0.5*(i + k - 1))) ...
                              + z(1 + round(0.5*(i - k - 1)), 1 + round(0.5*(i + k + 1))) ...
                              + z(1 + round(0.5*(i - k - 1)), 1 + round(0.5*(i + k - 1))));
                        dummy_z = dummy_z + sigma * 0.25*( ...
                                u(1 + round(0.5*(i - k + 1)), 1 + round(0.5*(i + k + 1))) ... 
                              + u(1 + round(0.5*(i - k + 1)), 1 + round(0.5*(i + k - 1))) ...
                              + u(1 + round(0.5*(i - k - 1)), 1 + round(0.5*(i + k + 1))) ...
                              + u(1 + round(0.5*(i - k - 1)), 1 + round(0.5*(i + k - 1))));
                    end;
                else
                    dummy_u = dummy_u + sigma * ...
                          z(1 + round(0.5*(i - k)), 1 + round(0.5*(i + k)));
                    dummy_z = dummy_z + sigma * ...
                          u(1 + round(0.5*(i - k)), 1 + round(0.5*(i + k)));
                end;
            end;

            I_u1(1, i + 1) = dummy_u;
            I_z1(1, i + 1) = dummy_z;
        end;

        % Create Toeplitz matrix
        for i = 1:N
            I_u1(i, i:N) = I_u1(1, 1:(N - i + 1)); % Faster than the Matlab command toepelitz()
            I_z1(i, i:N) = I_z1(1, 1:(N - i + 1));
        end;
%         toc;

        %% I3
%         tic;
        I_u2 = zeros(N);
        I_z2 = zeros(N);

        for i = 1:(N - 1)
            for j = 1:i
                I_u2(j + 1, i + 1) = I_u2(j, i) + 0.5*w(j, i) + 0.5*w(j + 1, i + 1);
                I_z2(j + 1, i + 1) = I_z2(j, i) + 0.5*v(j, i) + 0.5*v(j + 1, i + 1);
            end;
        end;
%         toc;

        %% Store values, and start over again
        u = qi*v0 + a*Delta*(0.5*qi*I_u1 + I_u2);
        v = v0 + 0.5*a*Delta*I_v1;
        w = w0 - 0.5*b*Delta*I_w1;
        z = q*w0 - b*Delta*(0.5*q*I_z1 +  I_z2);
    end;
    
    grid = kron((0:Delta:1), ones(N, 1));
    e_plus  = exp(0.5*gamma*grid);
    e_minus = exp(-0.5*gamma*grid);

    Laa = e_plus.*e_minus'.*u;
    Lab = e_plus.*e_plus'.*v;
    Lba = e_minus.*e_minus'.*w;
    Lbb = e_minus.*e_plus'.*z;