function [Puu, Puv, Pvu, Pvv, u, v, w, z] = P_kernel_solver(N, a, b, gamma, q, numIter)
%% [Kvu, Kvv] = K_kernel_solver_final(N, a, b, gamma, q, numIter)
% Solves a system on the form
%   Puu_x(x, xi) + Puu_xi(x, xi) =   a exp( gamma x) Pvu(x, xi)
%   Puv_x(x, xi) - Puv_xi(x, xi) =   a exp( gamma x) Pvv(x, xi)
%   Pvu_x(x, xi) - Pvu_xi(x, xi) = - b exp(-gamma x) Puu(x, xi)
%   Pvv_x(x, xi) + Pvv_xi(x, xi) = - b exp(-gamma x) Puv(x, xi)
%   
%   with  BCs:
%   Puu(0, xi) = q Pvu(0, xi)
%   Puv =   0.5 a exp( gamma x)
%   Pvu = - 0.5 b exp(-gamma x)
%   Pvv(0, xi) = (1/q) Puv(0, xi)
%
%   using the method of Characteristics with numIter iterations. The result
%   is returned in two N x N matrices.

    %% Implementation starts here
    mask = tril(ones(N));

    v0 = mask * (0.5*a);
    w0 = mask * (-0.5*b);
    
    Delta = 1/(N - 1);
    
    u = q*w0;
    v = v0;
    w = w0;
    z = (1/q)*v0;

    for iter = 1:numIter
        %% I1
%         tic;
        I_v1 = zeros(N);
        I_w1 = zeros(N);
    
        % Special treatment for the first diagonal line
        for j = 0:(N - 2)
            I_v1(j + 2, j + 1) = ...
                + 0.25*z(j + 1, j + 1) ...
                + 0.25*z(j + 2, j + 2) ...
                + 0.5*z(j + 2, j + 1);
            
            I_w1(j + 2, j + 1) = ...
                + 0.25*u(j + 1, j + 1) ...
                + 0.25*u(j + 2, j + 2) ...
                + 0.5*u(j + 2, j + 1);
        end;

        for j = 2:(N - 1)
            for i = 0:(N - 1 - j) 
                I_v1(j + i + 1, i + 1) = ...
                    + I_v1(j + i, i + 2) ...
                    + 0.25*z(j + i, i + 1) ... 
                    + 0.75*z(j + i + 1, i + 1) ...
                    + 0.75*z(j + i, i + 2) ...
                    + 0.25*z(j + i + 1, i + 2);
                
                I_w1(j + i + 1, i + 1) = ...
                    + I_w1(j + i, i + 2) ...
                    + 0.25*u(j + i, i + 1) ... 
                    + 0.75*u(j + i + 1, i + 1) ...
                    + 0.75*u(j + i, i + 2) ...
                    + 0.25*u(j + i + 1, i + 2);
            end;
        end;
%         toc;


        %% I2
%         tic;
        I_u1 = zeros(N);
        I_z1 = zeros(N);

        for j = 1:(N - 1)
            dummy_u = 0;
            dummy_z = 0;

            for k = 0:j
                if ((k == 0) || (k == j))
                    sigma = 0.5;
                else
                    sigma = 1;
                end;

                if (mod(j + k, 2))
                    if (k == 0)
                        dummy_u = dummy_u + sigma * 0.5*( ...
                                u(1 + round(0.5*(j + k + 1)), 1 + round(0.5*(j - k + 1))) ... 
                              + u(1 + round(0.5*(j + k - 1)), 1 + round(0.5*(j - k - 1))));
                        dummy_z = dummy_z + sigma * 0.5*( ...
                                z(1 + round(0.5*(j + k + 1)), 1 + round(0.5*(j - k + 1))) ... 
                              + z(1 + round(0.5*(j + k - 1)), 1 + round(0.5*(j - k - 1))));
                    else
                        dummy_u = dummy_u + sigma * 0.25*( ...
                                u(1 + round(0.5*(j + k + 1)), 1 + round(0.5*(j - k + 1))) ... 
                              + u(1 + round(0.5*(j + k + 1)), 1 + round(0.5*(j - k - 1))) ...
                              + u(1 + round(0.5*(j + k - 1)), 1 + round(0.5*(j - k + 1))) ...
                              + u(1 + round(0.5*(j + k - 1)), 1 + round(0.5*(j - k - 1))));
                        dummy_z = dummy_z + sigma * 0.25*( ...
                                z(1 + round(0.5*(j + k + 1)), 1 + round(0.5*(j - k + 1))) ... 
                              + z(1 + round(0.5*(j + k + 1)), 1 + round(0.5*(j - k - 1))) ...
                              + z(1 + round(0.5*(j + k - 1)), 1 + round(0.5*(j - k + 1))) ...
                              + z(1 + round(0.5*(j + k - 1)), 1 + round(0.5*(j - k - 1))));
                    end;
                else
                    dummy_u = dummy_u + sigma * ...
                          u(1 + round(0.5*(j + k)), 1 + round(0.5*(j - k)));
                    dummy_z = dummy_z + sigma * ...
                          z(1 + round(0.5*(j + k)), 1 + round(0.5*(j - k)));
                end;
            end;

            I_u1(j + 1, 1) = dummy_u;
            I_z1(j + 1, 1) = dummy_z;
        end;

        % Create Toeplitz matrix
        for j = 1:N
            I_u1(j:N, j) = I_u1(1:(N - j + 1), 1); % Faster than the Matlab command toepelitz()
            I_z1(j:N, j) = I_z1(1:(N - j + 1), 1);     
        end;
%         toc;

        %% I3
%         tic;
        I_u2 = zeros(N);
        I_z2 = zeros(N);

        for i = 1:(N - 1)
            for j = i:(N - 1)
                I_u2(j + 1, i + 1) = I_u2(j, i) + 0.5*w(j, i) + 0.5*w(j + 1, i + 1);
                I_z2(j + 1, i + 1) = I_z2(j, i) + 0.5*v(j, i) + 0.5*v(j + 1, i + 1);
            end;
        end;
%         toc;

        %% Store values, and start over again
        u = q*w0 + 0.5*q*b*Delta*I_u1 + a*Delta*I_u2;
        v = v0 - 0.5*a*Delta*I_v1;
        w = w0 + 0.5*b*Delta*I_w1;
        z = (1/q)*v0 - 0.5*(1/q)*a*Delta*I_z1 - b*Delta*I_z2;
    end;
    
    grid = kron((0:Delta:1), ones(N, 1));
    e_plus  = exp(0.5*gamma*grid);
    e_minus = exp(-0.5*gamma*grid);

    Puu = e_plus.*e_minus'.*u;
    Puv = e_plus.*e_plus'.*v;
    Pvu = e_minus.*e_minus'.*w;
    Pvv = e_minus.*e_plus'.*z;