function [num, den] = Laguerre_Gram(q, N, alpha, gamma, r)
    %% [num, den] = Laguerre_Gram(q, N, alpha, gamma, r)
    % 
    % Calculates the numerator and denominator of the best approximating
    %   polynomial of order r for a function which the N first Laguerre
    %   coefficients are passed as q.
    %   The method used is the one presented in [1].
    % 
    %   The associated Laguerre functions l_n(t; alpha, gamma) are related 
    %   to the Laguerre polynomials L_n(x) through:
    %   l_n(t; alpha, gamma) = \sqrt(gamma) * exp(-alpha*t) * L_n(gamma*t)
    %
    %   References:
    %       [1] A. Amghayrir, N. Tanguy, P. Brhonnet, P. Vilb and 
    %           L.-C. Calvez (2005): 
    %           "Laguerre-Gram Reduced-order Modeling"
    %               IEEE Transactions on Automatic Control 
    %               Vol 50. No 9, pp 1432-1435
    
    %% Find the coefficients a by construction of the matrix psi
    psi_aug = zeros(r + 1);

    for k = 0:r 
        psi_aug(1, k + 1) = 0;
        for n = 0:(N - k - 1)
             psi_aug(1, k + 1) = psi_aug(1, k + 1) + q(n + 1)*q(n + k + 1);
        end;
    end;

    for i = 1:(r - 1)
        for k = i:r
            psi_aug(i + 1, k + 1) = psi_aug(i, k) - q(i)*q(k);
        end;
    end;

    psi = psi_aug(1:r, 1:r);
    psi = psi + psi'.*(1 - eye(r)); %% Use symmetry to complete lower part
    b = psi_aug(1:r, r + 1);
    a = -psi\b;

    %% Construct polynomials
    a(r + 1) = 1; % A trick to ease the computation of the polynomials

    num = zeros(r, 1);
    for i = 1:r
        part = zeros(r, 1);
        for k = 0:(i-1)
            part = part + q(k + 1) * ...
                    conv(poly(-alpha*ones(i - k - 1, 1)), ...
                         poly((gamma - alpha)*ones(r - i + k, 1)))';
        end;

        num = num + a(i + 1)*part;
    end;
    num = num*sqrt(gamma);

    den = zeros(r+1, 1);
    for i = 0:r
        den = den + a(i + 1) * ...
                conv(poly(-alpha*ones(i, 1)), ...
                     poly((gamma - alpha)*ones(r - i, 1)))';
    end;

    %% Force the denominator monic
    num = num'./(den(1));
    den = den'./(den(1));